@@ -46,10 +46,14 @@ LOGPENDP((conn)->endp, cat, level, "CI:%s " fmt, \
 /* Specific rtp connection type (see struct mgcp_conn_rtp) */
 enum mgcp_conn_rtp_type {
 	MGCP_RTP_DEFAULT	= 0,
-	MGCP_OSMUX_BSC,
-	MGCP_OSMUX_BSC_NAT,
+	MGCP_RTP_OSMUX,
 	MGCP_RTP_IUUP,
 };
+extern const struct value_string mgcp_conn_rtp_type_names[];
+static inline const char *mgcp_conn_rtp_type_name(enum mgcp_conn_rtp_type val)
+{
+	return get_value_string(mgcp_conn_rtp_type_names, val);
+}
 
 /*! Connection type, specifies which member of the union "u" in mgcp_conn
  *  contains a useful connection description (currently only RTP) */
@@ -80,19 +84,20 @@ struct mgcp_conn_rtp {
 	struct {
 		/* Osmux state: disabled, activating, active */
 		enum osmux_state state;
-		/* Is cid holding valid data? is it allocated from pool? */
-		bool cid_allocated;
-		/* Allocated Osmux circuit ID for this conn */
-		uint8_t cid;
-		/* handle to batch messages */
+		/* Is local_cid holding valid data? is it allocated from pool? */
+		bool local_cid_allocated;
+		/* Allocated local Osmux circuit ID for this conn */
+		uint8_t local_cid;
+		/* Is remote_cid holding valid data? was it already received from client? */
+		bool remote_cid_present;
+		/* Received remote Osmux circuit ID for this conn */
+		uint8_t remote_cid;
+		/* handle to batch messages, shared (refcounted) among several conns */
 		struct osmux_in_handle *in;
-		/* handle to unbatch messages */
-		struct osmux_out_handle out;
-		/* statistics */
-		struct {
-			uint32_t chunks;
-			uint32_t octets;
-		} stats;
+		/* handle to unbatch messages, one allocated and owned per conn */
+		struct osmux_out_handle *out;
+		/* statistics: */
+		struct rate_ctr_group *ctrg;
 	} osmux;
 
 	struct {
@@ -103,7 +108,7 @@ struct mgcp_conn_rtp {
 		struct osmo_iuup_rnl_prim *init_ind;
 	} iuup;
 
-	struct rate_ctr_group *rate_ctr_group;
+	struct rate_ctr_group *ctrg;
 };
 
 /*! MGCP connection (untyped) */
@@ -145,12 +150,12 @@ struct mgcp_conn {
 enum {
 	IN_STREAM_ERR_TSTMP_CTR,
 	OUT_STREAM_ERR_TSTMP_CTR,
-        RTP_PACKETS_RX_CTR,
-        RTP_OCTETS_RX_CTR,
-        RTP_PACKETS_TX_CTR,
-        RTP_OCTETS_TX_CTR,
-        RTP_DROPPED_PACKETS_CTR,
-        RTP_NUM_CONNECTIONS,
+	RTP_PACKETS_RX_CTR,
+	RTP_OCTETS_RX_CTR,
+	RTP_PACKETS_TX_CTR,
+	RTP_OCTETS_TX_CTR,
+	RTP_DROPPED_PACKETS_CTR,
+	RTP_NUM_CONNECTIONS,
 };
 
 /* RTP per-connection statistics. Instances of the corresponding rate counter group
@@ -163,7 +168,7 @@ static const struct rate_ctr_desc mgcp_conn_rate_ctr_desc[] = {
 	[RTP_OCTETS_RX_CTR] = {"rtp:octets_rx", "Inbound rtp octets."},
 	[RTP_PACKETS_TX_CTR] = {"rtp:packets_tx", "Outbound rtp packets."},
 	[RTP_OCTETS_TX_CTR] = {"rtp:octets_tx", "Outbound rtp octets."},
-	[RTP_DROPPED_PACKETS_CTR] = {"rtp:dropped", "dropped rtp packets."}
+	[RTP_DROPPED_PACKETS_CTR] = {"rtp:dropped", "Dropped rtp packets."}
 };
 
 /* Aggregated RTP connection stats. These are updated when an RTP connection is freed.
@@ -181,9 +186,49 @@ static const struct rate_ctr_desc all_rtp_conn_rate_ctr_desc[] = {
 	[RTP_NUM_CONNECTIONS] = {"all_rtp:num_closed_conns", "Total number of rtp connections closed."}
 };
 
+/* Osmux connection related counters */
+enum {
+	OSMUX_CHUNKS_RX_CTR,
+	OSMUX_OCTETS_RX_CTR,
+	OSMUX_RTP_PACKETS_TX_CTR,
+	OSMUX_RTP_PACKETS_TX_DROPPED_CTR,
+	OSMUX_AMR_OCTETS_TX_CTR,
+	/* Only available in global stats: */
+	OSMUX_NUM_CONNECTIONS,
+	OSMUX_PACKETS_RX_CTR,
+	OSMUX_PACKETS_TX_CTR,
+	OSMUX_DROPPED_PACKETS_CTR,
+};
+
+/* RTP per-connection statistics. Instances of the corresponding rate counter group
+ * exist for the lifetime of an RTP connection.
+ * Must be kept in sync with all_rtp_conn_rate_ctr_desc below */
+static const struct rate_ctr_desc mgcp_conn_osmux_rate_ctr_desc[] = {
+	[OSMUX_CHUNKS_RX_CTR] = {"osmux:chunks_rx", "Inbound Osmux chunks."},
+	[OSMUX_OCTETS_RX_CTR] = {"osmux:octets_rx", "Inbound Osmux octets."},
+	[OSMUX_RTP_PACKETS_TX_CTR] = {"osmux:rtp_packets_tx", "Tx outbound RTP packets to encode as Osmux."},
+	[OSMUX_RTP_PACKETS_TX_DROPPED_CTR] = {"osmux:rtp_packets_tx_dropped", "Dropped Tx outbound RTP packets to encode as Osmux."},
+	[OSMUX_AMR_OCTETS_TX_CTR] = {"osmux:amr_octets_tx", "Tx outbound AMD payload octets."},
+};
+
+/* Aggregated Osmux connection stats. These are updated when an Osmux connection is freed.
+ * Must be kept in sync with mgcp_conn_osmux_rate_ctr_desc above */
+static const struct rate_ctr_desc all_osmux_conn_rate_ctr_desc[] = {
+	[OSMUX_CHUNKS_RX_CTR] = {"all_osmux:chunks_rx", "Inbound Osmux chunks."},
+	[OSMUX_OCTETS_RX_CTR] = {"all_osmux:octets_rx", "Inbound Osmux octets."},
+	[OSMUX_RTP_PACKETS_TX_CTR] = {"all_osmux:rtp_packets_tx", "Tx outbound RTP packets to encode as Osmux."},
+	[OSMUX_RTP_PACKETS_TX_DROPPED_CTR] = {"all_osmux:rtp_packets_tx_dropped", "Dropped Tx outbound RTP packets to encode as Osmux."},
+	[OSMUX_AMR_OCTETS_TX_CTR] = {"all_osmux:amr_octets_tx", "Tx outbound AMD payload octets."},
+	/* These last counters below do not exist in per-connection stats, only here: */
+	[OSMUX_NUM_CONNECTIONS] = {"all_osmux:num_closed_conns", "Total number of osmux connections closed."},
+	[OSMUX_PACKETS_RX_CTR] = {"all_osmux:packets_rx", "Total inbound UDP/Osmux packets."},
+	[OSMUX_PACKETS_TX_CTR] = {"all_osmux:packets_tx", "Total outbound UDP/Osmux packets."},
+	[OSMUX_DROPPED_PACKETS_CTR] = {"all_osmux:dropped_packets", "Dropped outbound UDP/Osmux packets."}
+};
+
 /* Was conn configured to handle Osmux? */
 static inline bool mgcp_conn_rtp_is_osmux(const struct mgcp_conn_rtp *conn) {
-	return conn->type == MGCP_OSMUX_BSC || conn->type == MGCP_OSMUX_BSC_NAT;
+	return conn->type == MGCP_RTP_OSMUX;
 }
 
 /* Was conn configured to handle Osmux? */
